;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Balloon Gliders
;  by Mandew
;
; Balloon Gliders will attach to Mario as he passes by them.
; Mario will glide upwards or downwards while using them.
; Press B or release Y to let 'em go.
;
; Uses Extra Bit: YES
;  Changes between 1-balloon glider and 2-balloon glider.
;   1-balloon glides downwards, 2-balloon glides upwards.
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;  Defines

; Customizables
!DipSpeed = #$38
!MaxDownSpeed = #$0C
!MaxUpSpeed = #$F0

; Sprite Tables

!CustomSpriteNumber = $7FAB9E
!CustomSpriteProperties = $7FAB10
!BehaviorState = $C2
!SpriteXPosLo = $E4
!SpriteXPosHi = $14E0
!SpriteXSpeed = $B6
!SpriteYPosLo = $D8
!SpriteYPosHi = $14D4
!SpriteYSpeed = $AA
!SpriteGFXProp = $15F6
!SpriteOffScreenVertical = $186C
!SpriteStatus = $14C8

; OAM
!OAM_X = $0300
!OAM_Y = $0301
!OAM_Tile = $0302
!OAM_Properties = $0303

; Mario

!MarioXPosLo = $94
!MarioXPosHi = $95
!MarioXSpeed = $7B
!MarioYPosLo = $96
!MarioYPosHi = $97
!MarioYSpeed = $7D
!MarioHoldingObject = $148F
!MarioAir = $72
!MarioSpinJump = $140D

; RAM Addresses
!LockFlag = $9D
!IOHeldButtons = $15
!IOPressedButtons = $16
!IOPressedButtons2 = $18
!SoundFXA = $1DFA
!SoundFXB = $1DFC
!EffectiveCounter = $14

; Routines

!JumpPointer = $0086DF
!GetMarioClipping = $03B664
!GetSpriteClipping = $03B69F
!CheckForCollisions = $03B72B
!UpdateXPos = $018022
!UpdateYPos = $01801A
!FinishOAMWrite = $01B7B3

; Constants

!AndRunButton = #$40
!JumpSound = #$01
!SpinJumpSound = #$04
!SpriteLoopInit = #$0C
!AndIsCustomSprite = #$02
!ExtraBits = #$04

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Wrappers
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

	print "INIT ",pc                                 
		RTL

	print "MAIN ",pc                                 
		PHB
		PHK
		PLB
		JSR MainRoutine
		PLB
		RTL

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; MainRoutine
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

	Return_Main:
		RTS

	MainRoutine:
		JSR SubOffscreenX3
		JSR DrawGraphics

; GameState Check
		LDA !SpriteStatus,x
		CMP #$08
		BNE Return_Main
		LDA !LockFlag
		BNE Return_Main ; Don't proceed if sprites are locked.

; Behavior Check
		LDA !BehaviorState,x
		JSL !JumpPointer

	BehaviorPointer:
		dw BehaviorIdle
		dw BehaviorGlide
		dw BehaviorAway

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; BehaviorIdle
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

	Return_Idle:
		RTS

	BehaviorIdle:
		JSL !GetMarioClipping
		JSL !GetSpriteClipping	
		JSL !CheckForCollisions
		BCC Return_Idle	   ; Don't Proceed unless there is a collision.

		JSR Sub_CancelActiveGliders
		JSR Sub_ActivateThisGlider
		
		RTS

; *************************

Sub_CancelActiveGliders:
	LDA !CustomSpriteNumber,x
	STA $00

	PHX
	LDX !SpriteLoopInit 	;Initiate a Sprite Loop

KeepLooping_A:
	DEX
	CPX #$00
	BMI StopLooping_A	; Don't proceed and stop looping if all the sprites were checked.

	LDA !CustomSpriteProperties,x
	AND !AndIsCustomSprite
	BNE KeepLooping_A	; Don't proceed unless sprite is custom; If not custom, go to KeepLooping_A

	LDA !CustomSpriteNumber,x
	CMP $00
	BNE KeepLooping_A	; Don't proceed unless the sprite's number is the same as this sprite's number.

	LDA !BehaviorState,x
	CMP #$01
	BNE KeepLooping_A	; Don't proceed unless the sprite is on BehaviorState #$01

	INC !BehaviorState,x	
	BRA KeepLooping_A	; Keep on looping.

StopLooping_A:
	PLX
	RTS

; *************************

Sub_ActivateThisGlider:
	STZ !MarioSpinJump
	INC !BehaviorState,x

	LDA !MarioYSpeed
	BMI NoDipSpeed
	CMP !DipSpeed
	BMI NoDipSpeed
	LDA !DipSpeed
	STA !SpriteYSpeed,x
	STA !MarioYSpeed

NoDipSpeed:
	LDA !SpriteYPosLo,x
	SEC
	SBC #$10
	STA !MarioYPosLo
	LDA !SpriteYPosHi,x
	SBC #$00
	STA !MarioYPosHi
	RTS

; *************************

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; BehaviorGlide
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

BehaviorGlide:
	JSR Sub_CheckIfStillRiding
	BCC CancelThisGlider
	BRA YouAreGliding

;***

CancelThisGlider:
	LDA !MarioXSpeed
	STA !SpriteXSpeed,x
	LDA !IOPressedButtons
	BMI MarioJumpsOff

	LDA !IOPressedButtons2
	BMI MarioSpinJumpsOff

MarioDropsOff:
	;STZ !MarioYSpeed	
	;INC !BehaviorState,x
	RTS
		
MarioSpinJumpsOff:
	LDA #$04
	STA !SoundFXB
	LDA #$01
	STA !MarioSpinJump
	BRA OffThisThing

MarioJumpsOff:
	LDA !JumpSound
	STA !SoundFXA

OffThisThing:
	LDA #$0B
	STA !MarioAir
	LDA #$C0
	STA !MarioYSpeed
	INC !BehaviorState,x
	RTS

;***

YouAreGliding:
	LDA !MarioXPosLo
	STA !SpriteXPosLo,x
	LDA !MarioXPosHi
	STA !SpriteXPosHi,x  ; Let the sprite copy Mario's X Position

	LDA !MarioYPosLo
	STA !SpriteYPosLo,x
	LDA !MarioYPosHi
	STA !SpriteYPosHi,x  ; Let the sprite copy Mario's Y Position

	LDA !MarioYSpeed
	STA !SpriteYSpeed,x

	LDA !CustomSpriteProperties,x
	AND !ExtraBits
	BNE BalloonGlide2   ; If it is a 2-balloon glider, go to 2BalloonGlide

BalloonGlide1:
	LDA !SpriteYSpeed,x
	BMI TooMuchSpeed1
	CMP !MaxDownSpeed
	BMI TooMuchSpeed1	

	JSR SuperBuyo
TooMuchSpeed1:
	STA !SpriteYSpeed,x
	STA !MarioYSpeed
	RTS

BalloonGlide2:
	LDA !MaxUpSpeed
	STA $00

	JSR SuperBuyo
	BPL NotSpeedLimit2
	CMP !MaxUpSpeed
	BPL NotSpeedLimit2

	LDA !MaxUpSpeed

NotSpeedLimit2:
	STA !SpriteYSpeed,x
	STA !MarioYSpeed
	RTS

DATA_UpdraftForce:
db $0A,$09,$08,$08,$07,$07,$06,$06
db $00,$00,$06,$06,$08,$08,$0A,$0C

SuperBuyo:
	LDA !MarioYSpeed
	PHA
	LSR
	LSR
	LSR
	LSR
	TAY
	PLA

	SEC
	SBC DATA_UpdraftForce,y
	STA !SpriteYSpeed,x
	STA !MarioYSpeed

	;LDA !SpriteYSpeed,x
	;SEC
	;SBC #$08
	;STA !SpriteYSpeed,x
	;STA !MarioYSpeed
	RTS

; *************************

Sub_CheckIfStillRiding:
	LDA !IOPressedButtons
	BMI ClearCarry
	
	LDA !IOPressedButtons2
	BMI ClearCarry

	;LDA !IOHeldButtons
	;AND !AndRunButton
	;BEQ ClearCarry

	SEC
	RTS
ClearCarry:
	CLC
	RTS

; *************************

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; BehaviorAway
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


BehaviorAway:
	LDA !SpriteYSpeed,x
	DEC
	DEC
	CMP #$A0
	BPL NotSpeedLimit3

	LDA #$A0

NotSpeedLimit3:
	STA !SpriteYSpeed,x

	JSL !UpdateXPos
	JSL !UpdateYPos
	RTS

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; DrawGraphics
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

GFX_XDisp: db $00,$00,$F4,$0C
	   db $00,$01,$F3,$0D
	   db $00,$00,$F4,$0C
	   db $00,$FF,$F5,$0B

GFX_YDisp: db $00,$E4,$E8,$E6
	   db $01,$E3,$E8,$E6
	   db $00,$E4,$E7,$E6
	   db $FF,$E4,$E8,$E5

GFX_Tile: db $00,$02,$02,$02
GFX_Property: db $04,$08,$06,$0A

DrawGraphics:
	JSR GET_DRAW_INFO

	LDA !EffectiveCounter	
	AND #$30
	LSR
	LSR
	STA $05

	LDA !SpriteGFXProp,x
	STA $04

	LDA !CustomSpriteProperties,x
	AND !ExtraBits
	BNE GFX_Entry2

GFX_Entry1:
	PHX
	LDA #$02
	TAX
	DEC
	STA $03
	BRA GraphicsLoop_A

GFX_Entry2:
	PHX
	LDA #$04
	TAX
	DEC
	STA $03

GraphicsLoop_A:
	DEX
	BMI StopGraphicsLoop_A

	PHX

	TXA
	CLC
	ADC $05
	TAX

	LDA $00
	CLC
	ADC GFX_XDisp,x
	STA !OAM_X,y

	LDA $01
	CLC
	ADC GFX_YDisp,x
	STA !OAM_Y,y
	
	PLX

	LDA GFX_Tile,x
	STA !OAM_Tile,y
	
	LDA $04
	AND #$01
	ORA GFX_Property,x
	ORA $64
	STA !OAM_Properties,y
	INY
	INY
	INY
	INY
	BRA GraphicsLoop_A

StopGraphicsLoop_A:
	PLX
	LDA $03              
        LDY #$02  
	JSL !FinishOAMWrite	
	RTS

;------------------------------------------------------------------------------------------
; Below are generic subroutines.
;------------------------------------------------------------------------------------------

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Get Draw Info
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

TABLE1:             db $0C,$1C
TABLE2:             db $01,$02

GET_DRAW_INFO:      STZ $186C,x             ; reset sprite offscreen flag, vertical
                    STZ $15A0,x             ; reset sprite offscreen flag, horizontal
                    LDA $E4,x               ; \
                    CMP $1A                 ;  | set horizontal offscreen if necessary
                    LDA $14E0,x             ;  |
                    SBC $1B                 ;  |
                    BEQ ON_SCREEN_X         ;  |
                    INC $15A0,x             ; /

ON_SCREEN_X:        LDA $14E0,x             ; \
                    XBA                     ;  |
                    LDA $E4,x               ;  |
                    REP #$20                ;  |
                    SEC                     ;  |
                    SBC $1A                 ;  | mark sprite invalid if far enough off screen
                    CLC                     ;  |
                    ADC #$0040            ;  |
                    CMP #$0180            ;  |
                    SEP #$20                ;  |
                    ROL A                   ;  |
                    AND #$01                ;  |
                    STA $15C4,x             ; / 
                    BNE INVALID             ; 
                    
                    LDY #$00                ; \ set up loop:
                    LDA $1662,x             ;  | 
                    AND #$20                ;  | if not smushed (1662 & 0x20), go through loop twice
                    BEQ ON_SCREEN_LOOP      ;  | else, go through loop once
                    INY                     ; / 
ON_SCREEN_LOOP:     LDA $D8,x               ; \ 
                    CLC                     ;  | set vertical offscreen if necessary
                    ADC TABLE1,y            ;  |
                    PHP                     ;  |
                    CMP $1C                 ;  | (vert screen boundry)
                    ROL $00                 ;  |
                    PLP                     ;  |
                    LDA $14D4,x             ;  | 
                    ADC #$00                ;  |
                    LSR $00                 ;  |
                    SBC $1D                 ;  |
                    BEQ ON_SCREEN_Y         ;  |
                    LDA $186C,x             ;  | (vert offscreen)
                    ORA TABLE2,y            ;  |
                    STA $186C,x             ;  |
ON_SCREEN_Y:        DEY                     ;  |
                    BPL ON_SCREEN_LOOP      ; /

                    LDY $15EA,x             ; get offset to sprite OAM
                    LDA $E4,x               ; \ 
                    SEC                     ;  | 
                    SBC $1A                 ;  | $00 = sprite x position relative to screen boarder
                    STA $00                 ; / 
                    LDA $D8,x               ; \ 
                    SEC                     ;  | 
                    SBC $1C                 ;  | $01 = sprite y position relative to screen boarder
                    STA $01                 ; / 
                    RTS                     ; return

INVALID:            PLA                     ; \ return from *main gfx routine* subroutine...
                    PLA                     ;  |    ...(not just this subroutine)
                    RTS                     ; /

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Off-screen
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

Table1:              db $0C,$1C
Table2:              db $01,$02
Table3:              db $40,$B0
Table6:              db $01,$FF
Table4:              db $30,$C0,$A0,$C0,$A0,$F0,$60,$90,$30,$C0,$A0,$80,$A0,$40,$60,$B0
Table5:              db $01,$FF,$01,$FF,$01,$FF,$01,$FF,$01,$FF,$01,$FF,$01,$00,$01,$FF

SubOffscreenX3:
LDA #$06

SubOffscreenMain:

STA $03

JSR SubIsOffscreen
BEQ Return2

LDA $5B
LSR
BCS VerticalLevel
LDA $D8,x
CLC
ADC #$50
LDA $14D4,x
ADC #$00
CMP #$02
BPL EraseSprite
LDA $167A,x
AND #$04
BNE Return2
LDA $13
AND #$01
ORA $03
STA $01
TAY
LDA $1A
CLC
ADC Table4,y
ROL $00
CMP $E4,x
PHP
LDA $1B
LSR $00
ADC Table5,y
PLP
SBC $14E0,x
STA $00
LSR $01
BCC Label20
EOR #$80
STA $00
Label20:
LDA $00
BPL Return2

EraseSprite:
LDA $14C8,x
CMP #$08
BCC KillSprite
LDY $161A,x
CPY #$FF
BEQ KillSprite
LDA #$00
STA $1938,y
KillSprite:
STZ $14C8,x
Return2:
RTS

VerticalLevel:

LDA $167A,x
AND #$04
BNE Return2
LDA $13
LSR
BCS Return2
AND #$01
STA $01
TAY
LDA $1C
CLC
ADC Table3,y
ROL $00
CMP $D8,x
PHP
LDA $1D
LSR $00
ADC Table6,y
PLP
SBC $14D4,x
STA $00
LDY $02
BEQ Label22
EOR #$80
STA $00
Label22:
LDA $00
BPL Return2
BMI EraseSprite

SubIsOffscreen:
LDA $15A0,x
ORA $186C,x
RTS